@@ -4,6 +4,7 @@ use Mojo::Base 'Mojo';
 # "Fry: Shut up and take my money!"
 use Carp ();
 use Mojo::Exception;
+use Mojo::Log;
 use Mojo::Util;
 use Mojolicious::Commands;
 use Mojolicious::Controller;
@@ -23,6 +24,19 @@ has commands => sub {
   return $commands;
 };
 has controller_class => 'Mojolicious::Controller';
+has log              => sub {
+  my $self = shift;
+
+  # Check if we have a log directory that is writable
+  my $log  = Mojo::Log->new;
+  my $home = $self->home;
+  my $mode = $self->mode;
+  $log->path($home->child('log', "$mode.log"))
+    if -d $home->child('log') && -w _;
+
+  # Reduced log output outside of development mode
+  return $mode eq 'development' ? $log : $log->level('info');
+};
 has mode => sub { $ENV{MOJO_MODE} || $ENV{PLACK_ENV} || 'development' };
 has moniker  => sub { Mojo::Util::decamelize ref shift };
 has plugins  => sub { Mojolicious::Plugins->new };
@@ -42,8 +56,8 @@ has static    => sub { Mojolicious::Static->new };
 has types     => sub { Mojolicious::Types->new };
 has validator => sub { Mojolicious::Validator->new };
 
-our $CODENAME = 'Clinking Beer Mugs';
-our $VERSION  = '6.15';
+our $CODENAME = 'Doughnut';
+our $VERSION  = '7.21';
 
 sub AUTOLOAD {
   my $self = shift;
@@ -65,7 +79,6 @@ sub build_controller {
   # Embedded application
   my $stash = {};
   if (my $sub = $tx->can('stash')) { ($stash, $tx) = ($tx->$sub, $tx->tx) }
-  $stash->{'mojo.secrets'} //= $self->secrets;
 
   # Build default controller
   my $defaults = $self->defaults;
@@ -128,16 +141,10 @@ sub handler {
 
   # Delayed response
   $self->log->debug('Nothing has been rendered, expecting delayed response')
-    unless $c->tx->is_writing;
+    unless $c->stash->{'mojo.rendered'};
 }
 
-sub helper {
-  my ($self, $name, $cb) = @_;
-  my $r = $self->renderer;
-  $self->log->debug(qq{Helper "$name" already exists, replacing})
-    if exists $r->helpers->{$name};
-  $r->add_helper($name => $cb);
-}
+sub helper { shift->renderer->add_helper(@_) }
 
 sub hook { shift->plugins->on(@_) }
 
@@ -145,8 +152,8 @@ sub new {
   my $self = shift->SUPER::new(@_);
 
   my $home = $self->home;
-  push @{$self->renderer->paths}, $home->rel_dir('templates');
-  push @{$self->static->paths},   $home->rel_dir('public');
+  push @{$self->renderer->paths}, $home->child('templates');
+  push @{$self->static->paths},   $home->child('public');
 
   # Default to controller and application namespace
   my $r = $self->routes->namespaces(["@{[ref $self]}::Controller", ref $self]);
@@ -158,20 +165,12 @@ sub new {
   $r->hide(qw(rendered req res respond_to send session signed_cookie stash));
   $r->hide(qw(tx url_for validation write write_chunk));
 
-  # Check if we have a log directory that is writable
-  my $mode = $self->mode;
-  $self->log->path($home->rel_file("log/$mode.log"))
-    if -d $home->rel_file('log') && -w _;
-
   $self->plugin($_)
     for qw(HeaderCondition DefaultHelpers TagHelpers EPLRenderer EPRenderer);
 
   # Exception handling should be first in chain
   $self->hook(around_dispatch => \&_exception);
 
-  # Reduced log output outside of development mode
-  $self->log->level('info') unless $mode eq 'development';
-
   $self->startup;
 
   return $self;
@@ -184,7 +183,7 @@ sub plugin {
 
 sub start {
   my $self = shift;
-  $_->_warmup for $self->static, $self->renderer;
+  $_->warmup for $self->static, $self->renderer;
   return $self->commands->run(@_ ? @_ : @ARGV);
 }
 
@@ -193,7 +192,7 @@ sub startup { }
 sub _exception {
   my ($next, $c) = @_;
   local $SIG{__DIE__}
-    = sub { ref $_[0] ? CORE::die($_[0]) : Mojo::Exception->throw(@_) };
+    = sub { ref $_[0] ? CORE::die $_[0] : Mojo::Exception->throw(shift) };
   $c->helpers->reply->exception($@) unless eval { $next->(); 1 };
 }
 
@@ -229,6 +228,12 @@ Mojolicious - Real-time web framework
 
 =head1 DESCRIPTION
 
+An amazing real-time web framework built on top of the powerful L<Mojo> web
+development toolkit. With support for RESTful routes, plugins, commands,
+Perl-ish templates, content negotiation, session management, form validation,
+testing framework, static file server, C<CGI>/C<PSGI> detection, first class
+Unicode support and much more for you to discover.
+
 Take a look at our excellent documentation in L<Mojolicious::Guides>!
 
 =head1 HOOKS
@@ -250,6 +255,26 @@ rather advanced features such as upload progress bars possible. Note that this
 hook will not work for embedded applications, because only the host application
 gets to build transactions. (Passed the transaction and application object)
 
+=head2 around_dispatch
+
+Emitted right after a new request has been received and wraps around the whole
+dispatch process, so you have to manually forward to the next hook if you want
+to continue the chain. Default exception handling with
+L<Mojolicious::Plugin::DefaultHelpers/"reply-E<gt>exception"> is the first hook
+in the chain and a call to L</"dispatch"> the last, yours will be in between.
+
+  $app->hook(around_dispatch => sub {
+    my ($next, $c) = @_;
+    ...
+    $next->();
+    ...
+  });
+
+This is a very powerful hook and should not be used lightly, it allows you to,
+for example, customize application-wide exception handling, consider it the
+sledgehammer in your toolbox. (Passed a callback leading to the next hook and
+the default controller object)
+
 =head2 before_dispatch
 
 Emitted right before the static file server and router start their work.
@@ -290,7 +315,7 @@ controller object)
 
 =head2 around_action
 
-Emitted right before an action gets invoked and wraps around it, so you have to
+Emitted right before an action gets executed and wraps around it, so you have to
 manually forward to the next hook if you want to continue the chain. Default
 action dispatching is the last hook in the chain, yours will run before it.
 
@@ -337,9 +362,9 @@ current controller object, a reference to the content and the format)
 
 =head2 after_dispatch
 
-Emitted in reverse order after a response has been rendered. Note that this
+Emitted in reverse order after a response has been generated. Note that this
 hook can trigger out of order due to its dynamic nature, and with embedded
-applications will only work for the application that is rendering.
+applications will only work for the application that is generating the response.
 
   $app->hook(after_dispatch => sub {
     my $c = shift;
@@ -349,26 +374,6 @@ applications will only work for the application that is rendering.
 Useful for rewriting outgoing responses and other post-processing tasks.
 (Passed the current controller object)
 
-=head2 around_dispatch
-
-Emitted right before the L</"before_dispatch"> hook and wraps around the whole
-dispatch process, so you have to manually forward to the next hook if you want
-to continue the chain. Default exception handling with
-L<Mojolicious::Plugin::DefaultHelpers/"reply-E<gt>exception"> is the first hook
-in the chain and a call to L</"dispatch"> the last, yours will be in between.
-
-  $app->hook(around_dispatch => sub {
-    my ($next, $c) = @_;
-    ...
-    $next->();
-    ...
-  });
-
-This is a very powerful hook and should not be used lightly, it allows you for
-example to customize application wide exception handling, consider it the
-sledgehammer in your toolbox. (Passed a callback leading to the next hook and
-the default controller object)
-
 =head1 ATTRIBUTES
 
 L<Mojolicious> inherits all attributes from L<Mojo> and implements the
@@ -394,16 +399,26 @@ Class to be used for the default controller, defaults to
 L<Mojolicious::Controller>. Note that this class needs to have already been
 loaded before the first request arrives.
 
+=head2 log
+
+  my $log = $app->log;
+  $app    = $app->log(Mojo::Log->new);
+
+The logging layer of your application, defaults to a L<Mojo::Log> object. The
+level will default to C<debug> if the L</mode> is C<development>, or C<info>
+otherwise. All messages will be written to C<STDERR>, or a C<log/$mode.log> file
+if a C<log> directory exists.
+
+  # Log debug message
+  $app->log->debug('It works');
+
 =head2 mode
 
   my $mode = $app->mode;
   $app     = $app->mode('production');
 
 The operating mode for your application, defaults to a value from the
-C<MOJO_MODE> and C<PLACK_ENV> environment variables or C<development>. Right
-before calling L</"startup">, L<Mojolicious> will pick up the current mode,
-name the log file after it and raise the log level from C<debug> to C<info> if
-it has a value other than C<development>.
+C<MOJO_MODE> and C<PLACK_ENV> environment variables or C<development>.
 
 =head2 moniker
 
@@ -437,6 +452,9 @@ L<Mojolicious::Guides::Rendering>.
   # Add another "templates" directory
   push @{$app->renderer->paths}, '/home/sri/templates';
 
+  # Add another "templates" directory with higher precedence
+  unshift @{$app->renderer->paths}, '/home/sri/themes/blue/templates';
+
   # Add another class with templates in DATA section
   push @{$app->renderer->classes}, 'Mojolicious::Plugin::Fun';
 
@@ -497,6 +515,9 @@ L<Mojolicious::Static> object.
   # Add another "public" directory
   push @{$app->static->paths}, '/home/sri/public';
 
+  # Add another "public" directory with higher precedence
+  unshift @{$app->static->paths}, '/home/sri/themes/blue/public';
+
   # Add another class with static files in DATA section
   push @{$app->static->classes}, 'Mojolicious::Plugin::Fun';
 
@@ -524,6 +545,12 @@ Validate values, defaults to a L<Mojolicious::Validator> object.
     return $value ne 'foo';
   });
 
+  # Add validation filter
+  $app->validator->add_filter(quotemeta => sub {
+    my ($validation, $name, $value) = @_;
+    return quotemeta $value;
+  });
+
 =head1 METHODS
 
 L<Mojolicious> inherits all methods from L<Mojo> and implements the following
@@ -550,8 +577,8 @@ Build L<Mojo::Transaction::HTTP> object and emit L</"after_build_tx"> hook.
 
   my $hash = $app->defaults;
   my $foo  = $app->defaults('foo');
-  $app     = $app->defaults({foo => 'bar'});
-  $app     = $app->defaults(foo => 'bar');
+  $app     = $app->defaults({foo => 'bar', baz => 23});
+  $app     = $app->defaults(foo => 'bar', baz => 23);
 
 Default values for L<Mojolicious::Controller/"stash">, assigned for every new
 request.
@@ -582,8 +609,10 @@ every request.
 
   $app->helper(foo => sub {...});
 
-Add a new helper that will be available as a method of the controller object
-and the application object, as well as a function in C<ep> templates.
+Add or replace a helper that will be available as a method of the controller
+object and the application object, as well as a function in C<ep> templates. For
+a full list of helpers that are available by default see
+L<Mojolicious::Plugin::DefaultHelpers> and L<Mojolicious::Plugin::TagHelpers>.
 
   # Helper
   $app->helper(cache => sub { state $cache = {} });
@@ -621,10 +650,9 @@ requests indiscriminately, for a full list of available hooks see L</"HOOKS">.
   my $app = Mojolicious->new({moniker => 'foo_bar'});
 
 Construct a new L<Mojolicious> application and call L</"startup">. Will
-automatically detect your home directory and set up logging based on your
-current operating mode. Also sets up the renderer, static file server, a
-default set of plugins and an L</"around_dispatch"> hook with the default
-exception handling.
+automatically detect your home directory. Also sets up the renderer, static file
+server, a default set of plugins and an L</"around_dispatch"> hook with the
+default exception handling.
 
 =head2 plugin
 
@@ -646,9 +674,9 @@ L<Mojolicious> distribution see L<Mojolicious::Plugins/"PLUGINS">.
   $app->start;
   $app->start(@ARGV);
 
-Start the command line interface for your application, for a full list of
-commands available by default see L<Mojolicious::Commands/"COMMANDS">. Note
-that the options C<-h>/C<--help>, C<--home> and C<-m>/C<--mode>, which are
+Start the command line interface for your application. For a full list of
+commands that are available by default see L<Mojolicious::Commands/"COMMANDS">.
+Note that the options C<-h>/C<--help>, C<--home> and C<-m>/C<--mode>, which are
 shared by all commands, will be parsed from C<@ARGV> during compile time.
 
   # Always start daemon
@@ -688,20 +716,20 @@ that have been bundled for internal use.
 
 =head2 Mojolicious Artwork
 
-  Copyright (C) 2010-2015, Sebastian Riedel.
+  Copyright (C) 2010-2017, Sebastian Riedel.
 
 Licensed under the CC-SA License, Version 4.0
 L<http://creativecommons.org/licenses/by-sa/4.0>.
 
 =head2 jQuery
 
-  Copyright (C) 2005, 2014 jQuery Foundation, Inc.
+  Copyright (C) jQuery Foundation.
 
 Licensed under the MIT License, L<http://creativecommons.org/licenses/MIT>.
 
 =head2 prettify.js
 
-  Copyright (C) 2006, 2013 Google Inc.
+  Copyright (C) 2006, 2013 Google Inc..
 
 Licensed under the Apache License, Version 2.0
 L<http://www.apache.org/licenses/LICENSE-2.0>.
@@ -711,6 +739,8 @@ L<http://www.apache.org/licenses/LICENSE-2.0>.
 Every major release of L<Mojolicious> has a code name, these are the ones that
 have been used in the past.
 
+7.0, C<Doughnut> (U+1F369)
+
 6.0, C<Clinking Beer Mugs> (U+1F37B)
 
 5.0, C<Tiger Face> (U+1F42F)
@@ -729,12 +759,6 @@ have been used in the past.
 
 1.0, C<Snowflake> (U+2744)
 
-0.999930, C<Hot Beverage> (U+2615)
-
-0.999927, C<Comet> (U+2604)
-
-0.999920, C<Snowman> (U+2603)
-
 =head1 SPONSORS
 
 Some of the work on this distribution has been sponsored by
@@ -752,8 +776,6 @@ Current members of the core team in alphabetical order:
 
 Abhijit Menon-Sen, C<ams@cpan.org>
 
-Dan Book, C<dbook@cpan.org>
-
 Glen Hinkle, C<tempire@cpan.org>
 
 Jan Henning Thorsen, C<jhthorsen@cpan.org>
@@ -784,6 +806,8 @@ Alexey Likhatskiy
 
 Anatoly Sharifulin
 
+Andre Parker
+
 Andre Vieth
 
 Andreas Jaekel
@@ -792,6 +816,8 @@ Andreas Koenig
 
 Andrew Fresh
 
+Andrew Nugged
+
 Andrey Khozov
 
 Andrey Kuzmin
@@ -834,6 +860,8 @@ chromatic
 
 Curt Tilmes
 
+Dan Book
+
 Daniel Kimsey
 
 Danijel Tasov
@@ -856,6 +884,8 @@ Dominique Dumont
 
 Douglas Christopher Wilson
 
+Eugen Konkov
+
 Eugene Toropov
 
 Gisle Aas
@@ -900,8 +930,12 @@ Kitamura Akatsuki
 
 Klaus S. Madsen
 
+Knut Arne Bjorndal
+
 Lars Balker Rasmussen
 
+Lee Johnson
+
 Leon Brocard
 
 Magnus Holm
@@ -944,6 +978,8 @@ Nils Diewald
 
 Oleg Zhelo
 
+Olivier Mengue
+
 Pascal Gaudette
 
 Paul Evans
@@ -976,6 +1012,8 @@ Robin Lee
 
 Roland Lammel
 
+Roy Storey
+
 Ryan Jendoubi
 
 Sascha Kiefer
@@ -1034,7 +1072,7 @@ Zoffix Znet
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2008-2015, Sebastian Riedel.
+Copyright (C) 2008-2017, Sebastian Riedel and others.
 
 This program is free software, you can redistribute it and/or modify it under
 the terms of the Artistic License version 2.0.
@@ -1042,6 +1080,6 @@ the terms of the Artistic License version 2.0.
 =head1 SEE ALSO
 
 L<https://github.com/kraih/mojo>, L<Mojolicious::Guides>,
-L<http://mojolicio.us>.
+L<http://mojolicious.org>.
 
 =cut
